(**************************************************************************)
(*                                                                        *)
(* Proof of the Heapsort Algorithm.                                       *)
(*                                                                        *)
(* Jean-Christophe Fillitre (LRI, Universit Paris Sud)                  *)
(* March 1999                                                             *)
(*                                                                        *)
(**************************************************************************)

Require ZArith.
Require Arrays.
Require Programs.
Require Exchange.
Require Permut.
Require Omega.

Require Swap.
Require Heap.
Require Inftree.

Lemma R11 : (k:Z) `2*k+1+1` = `2*k+2`.
Proof.
Intro; Omega.
Save.

(* To annotate the recursive function downheap, it is convenient to
 * introduce the following predicate, which expresses that j is the
 * greatest son of k. *)

Implicit Arguments On.

Inductive select_son [N:Z; t:(array N Z); k,n,j:Z] : Prop :=
    select_left_son : 
      `j = 2*k+1` -> (`2*k+2 <= n` -> (Zge t#[j] t#[`2*k+2`])) 
        -> (select_son t k n j)
  | select_right_son :
      `j = 2*k+2` -> `j <= n` -> (Zge t#[j] t#[`2*k+1`]) 
        -> (select_son t k n j).

Implicit Arguments Off.

(* The correctness of downheap requires the two following lemmas *)

Lemma Lemma_1 : (N:Z)(t0,t1,t2:(array N Z))
    (n0,k0,j':Z)
    `2*k0+1 <= n0`
 -> (select_son t0 k0 n0 j')
 -> `(access   t0 k0) < (access   t0 j')`
 -> (exchange t1 t0 k0 j')
 -> `0 <= k0 <= n0`
 -> `n0 < N`
 -> ((i:Z)`k0+1 <= i <= n0`->(heap t0 n0 i))
 -> ((i:Z)`j' <= i <= n0`->(heap t2 n0 i))
 -> ((i:Z)
        `0 <= i < j'`\/`j' < i < 2*j'+1`\/`n0 < i < N`
        ->`(access   t2 i) = (access   t1 i)`)
 -> ((v:Z)(inftree t1 n0 v j')->(inftree t2 n0 v j'))
 -> (i:Z)`k0 < i < j'`->(heap t2 n0 i).
Proof.
Intros. Apply heap_cons.
Elim H0; Omega'.
(* branch 2i+1 *)
Intro. Rewrite (H7 i). Rewrite (H7 `2*i+1`).
Decompose [exchange] H2. Rewrite (H16 i). Rewrite (H16 `2*i+1`).
Generalize H10. Elim (H5 i); Intros. Exact (H17 H23).
Omega'. Omega'. Omega'.
Elim H0; Omega'. Omega'. Omega'. Omega'.
Elim H0; Omega'. Elim H0; Omega'. 
Intro. Apply H6; Elim H0; Omega'.
(* branch 2i+2 *)
Intro. Rewrite (H7 i). Rewrite (H7 `2*i+2`).
Decompose [exchange] H2. Rewrite (H16 i). Rewrite (H16 `2*i+2`).
Generalize H10. Elim (H5 i); Intros. Exact (H20 H23).
Omega'. Omega'. Omega'. 
Elim H0; Omega'. Omega'. Omega'. Omega'.
Elim H0; Omega'. Elim H0; Omega'. 
Intro. Apply H6; Elim H0; Omega'.
Save.

Lemma Lemma_2 : (N:Z)(t0,t1,t2:(array N Z))
    (n0,k0,j':Z)
    `2*k0+1 <= n0`
 -> (select_son t0 k0 n0 j')
 -> `(access   t0 k0) < (access   t0 j')`
 -> (exchange t1 t0 k0 j')
 -> `0 <= k0 <= n0`
 -> `n0 < N`
 -> ((i:Z)`k0+1 <= i <= n0`->(heap t0 n0 i))
 -> ((i:Z)`j' <= i <= n0`->(heap t2 n0 i))
 -> ((i:Z)
        `0 <= i < j'`\/`j' < i < 2*j'+1`\/`n0 < i < N`
        ->`(access   t2 i) = (access   t1 i)`)
 -> ((v:Z)(inftree t1 n0 v j')->(inftree t2 n0 v j'))
 -> (i:Z)`k0 <= i <= n0`->(heap t2 n0 i).
Proof.
Intros.
Elim (Z_lt_ge_dec i j'); Intro HHi.
Elim (Z_le_lt_eq_dec k0 i); [ Intro HHHi | Intro HHHi | Intuition ].

(* 1. k0 < i < j' *)
Apply (Lemma_1 N t0 t1 t2 n0 k0 j'); Assumption Orelse Omega'.

(* 2. k0 = i *)
Apply heap_cons.
Omega'.
(* branch 2i+1 *)
(* t[k] >= t[2k+1] *)
Intro. Elim H0; Intros.
  (* j' = 2k+1 *)
  Rewrite <- HHHi. Rewrite <- H11.
  Rewrite (H7 k0). Decompose [exchange] H2. Rewrite H16.
  Apply Zle_ge. Apply inftree_1 with n:=n0.
  Apply H8.
  Apply inftree_2 with t1:=t0 k:=k0. Omega'. 
  Apply inftree_3.
  Apply H5; Omega'. Assumption. Omega'. Omega'. Omega'.
  (* j' = 2k+2 *)
  Rewrite <- HHHi.
  Rewrite (H7 k0). Decompose [exchange] H2. Rewrite H17.
  Rewrite (H7 `2*k0+1`). Rewrite (H19 `2*k0+1`).
  Omega'. Omega'. Omega'. Omega'. Omega'. Omega'. 
(* (heap t2 n (2k+1)) *)
Intro. Elim H0; Intros.
  (* j' = 2k+1 *)
  Apply H6; Omega'.
  (* j' = 2k+2 *)
  Apply (Lemma_1 N t0 t1 t2 n0 k0 j'); Assumption Orelse Omega'.
(* branch 2i+2 *)
(* t[k] >= t[2k+2] *)
Intro. Elim H0; Intros.
  (* j' = 2k+1 *)
  Rewrite <- HHHi.
  Rewrite (H7 k0). Decompose [exchange] H2. Rewrite H16.
  Rewrite (H7 `2*k0+2`). Rewrite (H18 `2*k0+2`).
  Omega'. Omega'. Omega'. Omega'. Omega'. Omega'. 
  (* j' = 2k+2 *)
  Rewrite <- HHHi. Rewrite <- H11.
  Rewrite (H7 k0). Decompose [exchange] H2. Rewrite H17.
  Apply Zle_ge. Apply inftree_1 with n:=n0.
  Apply H8.
  Apply inftree_2 with t1:=t0 k:=k0. Omega'. 
  Apply inftree_3.
  Apply H5; Omega'. Assumption. Omega'. Omega'. Omega'.
(* (heap t2 n (2k+2)) *)
Intro. Elim H0; Intros.
  (* j' = 2k+1 *)
  Apply H6; Omega'.
  (* j' = 2k+2 *)
  Apply H6; Omega'.

(* 3. i >= j' *)
Apply H6; Omega'.
Save.


(* The recursive function downheap *)

Correctness downheap
  let rec downheap (N:Z)(t:array N of Z)(k,n:Z) : unit { variant `n-k` } =
    {    `0 <= k <= n` /\ `n < N`
      /\ (i:Z) `k+1 <= i <= n` -> (heap t n i) }
    (let j = 2*k+1 in
     if j <= n then
       let j' = (if j+1 <= n then if t[j] < t[j+1] then j+1 else j else j)
                { (select_son t k n result) } in
       if t[k] < t[j'] then begin (swap N t k j'); (downheap N t j' n) end)
    {    (permut t t@)
      /\ ((i:Z) `k <= i <= n` -> (heap t n i))
      /\ ((i:Z) (`0<=i<k` \/ `k<i<2*k+1` \/ `n<i<N`) -> t#[i]=t@#[i])
      /\ ((v:Z) (inftree t@ n v k) -> (inftree t n v k)) }
.
Proof.
Omega'.
Omega'.

(* select_son: case 2k+1 <= n, 2k+2 <= n, t[2k+1] < t[2k+2] *)
Rewrite (R11 k0).
Apply select_right_son; 
  [ Reflexivity | Omega' | Rewrite (R11 k0) in Test1; Omega' ].
(* select_son: case 2k+1 <= n, 2k+2 <= n, t[2k+1] >= t[2k+2] *)
Apply select_left_son; 
  [ Reflexivity | Rewrite (R11 k0) in Test1; Intro; Assumption ].
(* select_son: case 2k+1 <= n, 2k+2 > n *)
Apply select_left_son; [ Reflexivity | Intro; Absurd `2*k0+2 <= n0`; Omega' ].

(* 0 <= j' < N *)
Elim Post16; Omega'.

(* 0 <= k0 < N0 *)
Omega'.

(* `0 <= k0 < N0`/\`0 <= j' < N0` *)
Split ; [ Omega' | Elim Post16; Omega' ].

(* (Zwf `0` `n0-j'` `n0-k0`) *)
Unfold Zwf; Elim Post16; Omega'.

(* pre-condition of recursive call *)
Split ; [ Elim Post16; Omega' | Split ; [ Omega' | Idtac ] ].
Intros i Hi.
Apply heap_id with t := t0.
Decompose [and] Pre5.
Apply H3; Elim Post16; Omega'.
Elim Post12; Intros.
Unfold array_id. Intros i' Hi'.
Symmetry; Apply (H3 i'); Elim Post16; Omega'.

(* post-condition: case of recursive call *)
Split; [ Idtac | Split; [ Idtac | Split ] ].
(* permut *)
Apply permut_trans with t' := t1.
Intuition. Apply exchange_is_permut with i:=k0 j:=j'; Assumption.
(* heap *)
Decompose [and] Pre5; Clear Pre5.
Decompose [and] Post11; Clear Post11.
Intros i Hi.
Apply (Lemma_2 N0 t0 t1 t2 n0 k0 j'); Assumption Orelse Omega'.
(* unchanged parts of the array *)
Intros i Hi.
Decompose [and] Post11; Clear Post11.
Rewrite (H1 i).
Decompose [exchange] Post12. 
Elim Post16; Intros. Apply H8; Omega'. Apply H8; Omega'.
Elim Post16; Intros; Omega'.
(* inftree *)
Decompose [and] Post11; Clear Post11.
Decompose [and] Pre5; Clear Pre5.
Intros v Hv.
Apply inftree_cons.
Split; Assumption.
Rewrite (H1 k0). 
Decompose [exchange] Post12. Rewrite H10. 
Elim Post16; Intros.
  (* j' = 2k+1 *)
  Rewrite H5. Generalize Test4. Case Hv; Intros.
  Apply inftree_1 with n:=n0. Auto.
  (* j' = 2k+2 *)
  Generalize H13. Rewrite H5. Case Hv; Intros.
  Apply inftree_1 with n:=n0. Auto.
Elim Post16; Intros; Omega'.
  (* branch 2k+1 *)
  Intro. Elim Post16; Intros.
    (* j' = 2k+1 *)
    Rewrite <- H8. Apply H3. 
    Apply inftree_2 with t1:=t0 k:=k0. Omega'. 
    Rewrite H8. Generalize H5. Case Hv; Auto.
    Assumption. Omega'. Omega'.
    (* j' = 2k+2 *)
    Apply inftree_trans with v:=t2#[`2*k0+1`].
    Rewrite (H1 `2*k0+1`).
    Decompose [exchange] Post12. Rewrite (H16 `2*k0+1`).
    Generalize H5. Case Hv; Intros.
    Apply inftree_1 with n:=n0. Auto.
    Omega'. Omega'. Omega'. Omega'.
    Apply inftree_3.
    Apply (Lemma_2 N0 t0 t1 t2 n0 k0 j'); Assumption Orelse Omega'.
  (* branch 2k+2 *)
  Intro. Elim Post16; Intros.
    (* j' = 2k+1 *)
    Apply inftree_trans with v:=t2#[`2*k0+2`].
    Rewrite (H1 `2*k0+2`).
    Decompose [exchange] Post12. Rewrite (H15 `2*k0+2`).
    Generalize H5. Case Hv; Intros.
    Apply inftree_1 with n:=n0. Auto.
    Omega'. Omega'. Omega'. Omega'.
    Apply inftree_3.
    Apply H; Omega'.
    (* j' = 2k+2 *)
    Rewrite <- H8. Apply H3. 
    Apply inftree_2 with t1:=t0 k:=k0. Omega'. 
    Rewrite H8. Generalize H5. Case Hv; Auto.
    Assumption. Omega'. Omega'.
    
(* t[k] >= t[j'] *)
Split; [ Idtac | Split; [ Idtac | Split ] ].
Auto with datatypes.
Intros i Hi. 
Decompose [and] Pre5. Clear Pre5.
Elim (Z_le_lt_eq_dec k0 i); [ Intro HHHi | Intro HHHi | Intuition ].
(* k0 < i *)
Apply H3; Omega'.
(* k0 = i *)
Rewrite <- HHHi. Apply heap_cons.
Omega'.
Intro. Elim Post16; Intros.
Rewrite <- H4. Assumption. Omega'.
Intro. Apply H3; Omega'.
Intro. Elim Post16; Intros.
Omega'. Rewrite <- H4. Assumption.
Intro. Apply H3; Omega'.
Auto.
Auto.

(* 2k+1 > n *)
Split; [ Idtac | Split; [ Idtac | Split ] ].
Auto with datatypes.
Intros i Hi. 
Decompose [and] Pre5. Clear Pre5.
Elim (Z_le_lt_eq_dec k0 i); [ Intro HHHi | Intro HHHi | Intuition ].
Apply H3; Omega'.
Rewrite <- HHHi. Apply heap_cons.
Omega'.
Intro. Absurd `2*k0+1 > n0`; Omega'.
Intro. Absurd `2*k0+1 > n0`; Omega'.
Intro. Absurd `2*k0+2 > n0`; Omega'.
Intro. Absurd `2*k0+2 > n0`; Omega'.
Auto.
Auto.
Save.
