(**************************************************************************)
(*                                                                        *)
(* Proof of the Heapsort Algorithm.                                       *)
(*                                                                        *)
(* Jean-Christophe Fillitre (LRI, Universit Paris Sud)                  *)
(* March 1999                                                             *)
(*                                                                        *)
(**************************************************************************)

Require ZArithRing.
Require Arrays.
Require Programs.
Require Exchange.
Require Permut.
Require Sorted.
Require Omega.

Require Swap.
Require Heap.
Require Inftree.
Require Downheap.

Lemma double_div2 : (x:Z) `(Zdiv2 (2*x)) = x`.
Proof.
Destruct x; Auto.
Save.

Lemma double_div2_bis : (x:Z) `0 <= x` -> `(Zdiv2 (2*x+1)) = x`.
Proof.
Destruct x; Auto.
Intros.
Simpl in H.
Absurd `0 <= (NEG p)`.
Simpl. Compute. Auto. Assumption.
Save.

Lemma lem_div2_0 : (n:Z)`1 <= n` -> `(-1) <= (Zdiv2 (n-2)) <= n-1`.
Proof.
Intros.
Elim (Z_lt_ge_dec `1` n); Intro.
Elim (Z_modulo_2 n).
Intro H0; Elim H0; Clear H0; Intros.
Replace `n-2` with `2*(x-1)`.
Rewrite double_div2.
Omega. Omega.
Intro H0; Elim H0; Clear H0; Intros.
Replace `n-2` with `2*(x-1)+1`.
Rewrite double_div2_bis.
Omega. Omega. Omega. Omega.
Replace n with `1`.
Simpl. Omega. Omega.
Save.

Lemma lem_div2_1 : (n:Z)`1 <= n` -> `0 <= (Zdiv2 (n-2))+1`.
Proof.
Intros n Hn. Generalize (lem_div2_0 n Hn). Omega.
Save.

Lemma lem_div2_2 : (n,i:Z)
  `1 <= n` -> `(Zdiv2 (n-2))+1 <= i <= n-1` -> `2*i >= n-1`.
Proof.
Intros n i Hn.
Elim (Z_lt_ge_dec `1` n); Intro.
Elim (Z_modulo_2 n).
Intro H0; Elim H0; Clear H0; Intros x Hx.
Replace `n-2` with `2*(x-1)`.
Rewrite double_div2.
Omega. Omega. 
Intro H0; Elim H0; Clear H0; Intros x Hx.
Replace `n-2` with `2*(x-1)+1`.
Rewrite double_div2_bis.
Omega. Omega. Omega. Omega.
Replace n with `1`.
Simpl. Omega. Omega.
Save.

Correctness heapsort 
  fun (N:Z)(t:array N of Z) ->
    { `1 <= N` }
    begin
      let k = ref (Zdiv2 (N-2)) in
      while !k >= 0 do
        { invariant `-1 <= k <= N-1` 
                 /\ ((i:Z) `k+1 <= i <= N-1` -> (heap t `N-1` i))
		 /\ (permut t t@0)
          variant `k+1` }
      	(downheap N t !k (N-1));
	k := !k-1
      done;
      assert { (heap t `N-1` `0`) /\ (permut t t@0) };
      let k = ref (N-1) in
      while !k >= 1 do
        { invariant `0 <= k <= N-1` 
                 /\ ((i:Z) `0 <= i <= k` -> (heap t k i))
		 /\ (`k+1 <= N-1` -> `(access t 0) <= (access t (k+1))`)
		 /\ (`k+1 <= N-1` -> (sorted_array t `k+1` `N-1`))
		 /\ (permut t t@0)	
          variant k }
        (swap N t 0 !k);
        (downheap N t 0 (!k-1));
	k := !k-1
      done
    end
    { (sorted_array t `0` `N-1`) /\ (permut t t@) }
.
Proof.
Split; [ Omega' | Split; [ Omega' | Intuition ] ]. 

(* preservation of first invariant *)
Split; [ Unfold Zwf; Omega' | Split ; [ Omega' | Split ] ].
Ring `k0-1+1`. Intuition.
Decompose [and] Post3; Clear Post3.
Decompose [and] Pre2; Clear Pre2.
Apply permut_trans with t':=t0; Assumption.

Decompose [and] Pre2; Clear Pre2.
Split; [ Apply H0; Omega' | Assumption ].

Split; [ Idtac | Split ]. 
Exact (lem_div2_0 N Pre6).
Intros i Hi; Apply heap_leaf.
Generalize (lem_div2_1 N Pre6); Omega'.
Exact (lem_div2_2 N i Pre6 Hi).
Auto with datatypes.

Omega'.

(* pre-condition of (downheap N t 0 (!k-1)) *)
Split; [ Omega' | Split; [ Omega' | Idtac ] ]. 
Intros i Hi.
Apply heap_id with t:=t1.
Decompose [and] Pre4; Clear Pre4.
Apply heap_weakening. Omega'.
Apply H0; Omega'. Omega'.
Decompose [exchange] Post9; Clear Post9.
Unfold array_id.
Intros i0 Hi0. Symmetry. Apply H4; Omega'.

(* preservation of second invariant *)
Decompose [and] Post8; Clear Post8.
Split; [ Unfold Zwf; Omega' 
       | Split; [ Omega' | Split; [ Idtac | Split; [ Idtac | Split ] ] ] ].
(* heap *)
Intros i Hi. Apply H; Omega'.
(* t[0] <= t[k] *)
Ring `k0-1+1`. Intro. 
Rewrite (H1 k0); [ Idtac | Omega' ].
Decompose [exchange] Post9.
Rewrite H8.
Apply inftree_1 with n:=`k0-1`.
Apply H3.
Apply inftree_weakening. Omega'.
Apply inftree_exchange with t1:=t1. Omega'.
Apply inftree_3.
Decompose [and] Pre4; Clear Pre4.
Apply H10; Omega'.
Assumption. Omega'.
(* sorted *)
Decompose [and] Pre4; Clear Pre4.
Ring `k0-1+1`. Intro. 
Elim (Z_le_lt_eq_dec k0 `N-1` H8); Intro.
  (* k0 < N-1 *)
  Replace k0 with `(k0+1)-1`; [ Idtac | Omega' ].
  Apply left_extension. Omega'. Omega'.
  Apply sorted_array_id with t1:=t2. 
  Apply sorted_array_id with t1:=t1. 
  Apply H7; Omega'.
  Decompose [exchange] Post9.
  Unfold array_id. Intros i Hi. Symmetry. Apply H15; Omega'.
  Unfold array_id. Intros i Hi. Symmetry. Apply H1; Omega'.
  (* t3[k0] <= t3[k0+1] *)
  Ring `k0+1-1`. 
  Rewrite (H1 k0); [ Idtac | Omega' ].
  Rewrite (H1 `k0+1`); [ Idtac | Omega' ].
  Decompose [exchange] Post9.
  Rewrite H14. Rewrite (H15 `k0+1`); [ Idtac | Omega' | Omega' | Omega' ].
  Apply H5; Omega'.
  (* k0 = N-1 *)
  Rewrite y. 
  Unfold sorted_array.
  Intros HN x HHx Hx. Absurd `x >= N-1`; Omega'.
(* (permut t3 t) *)
Decompose [and] Pre4; Clear Pre4.
Decompose [and] Pre5; Clear Pre5.
Apply permut_trans with t':=t2; Try Assumption. 
Apply permut_trans with t':=t1.
Apply exchange_is_permut with i:=`0` j:=k0. Assumption. 
Assumption. 

(* post-condition *)
Decompose [and] Pre4; Clear Pre4.
Split; [ Idtac | Assumption ].
Elim (Z_le_lt_eq_dec `1` N Pre6); Intro.
  (* 1 < N *)
  Replace `0` with `1-1`; [ Idtac | Omega' ].
  Apply left_extension. Omega'. Omega'.
  Replace `1` with `k0+1`; [ Idtac | Omega' ].
  Replace `N-(k0+1)` with `N-1`; [ Idtac | Omega' ].
  Apply H3; Omega'.
  Ring `1-1`.  Replace `1` with `k0+1`; [ Idtac | Omega' ].
  Apply H1; Omega'.
  (* 1 = N *)
  Unfold sorted_array. 
  Intros HN x HHx Hx. Absurd `x >= N-1`; Omega'.

Split; [ Omega' | Split; [ Idtac | Split; [ Idtac | Split ] ] ].
Decompose [and] Pre5.
Exact (heap_all H0).
Intro; Absurd `N-1+1 <= N-1`; Omega'.
Intro; Absurd `N-1+1 <= N-1`; Omega'.
Intuition.
Save.
